/*
                        Compile
    Linux:
        gcc ./fwupd.c -I /usr/include/PCSC -l pcsclite -o fwupd

    macOS:
        gcc ./fwupd.c -framework PCSC -o fwupd

    macOS Sonoma driver workaround:
        sudo defaults write /Library/Preferences/com.apple.security.smartcard useIFDCCID -bool yes

    For Windows x86_64
        x86_64-w64-mingw32-gcc ./fwupd.c -l winscard -o fwupd.exe

    For Windows i686
        i686-w64-mingw32-gcc ./fwupd.c -l winscard -o fwupd.exe
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifdef __MINGW32__
#include <winscard.h>
#else
#include <PCSC/wintypes.h>
#include <PCSC/winscard.h>
#endif

#define EUICC_INTERFACE_BUFSZ 264

#define FWUPD_BLOCK_SIZE 2061
#define FWUPD_TRANS_BLOCK_SIZE 255

#define APDU_FWUPD_SELECT_AID "\x01\xA4\x04\x00\x10\xA0\x65\x73\x74\x6B\x6D\x65\xFF\xFF\xFF\xFF\x66\x77\x75\x70\x64"
#define APDU_FWUPD_TRIGGER "\x01\x55\x55\x55\x00"

#define APDU_EXIT "\xAA\x00\x00\x00\x00"

#define APDU_BLOCK_WRITE "\xAA\x11\x00\x00\x00"
#define APDU_BLOCK_READ "\xAA\x12\x00\x00\x00"
#define APDU_BLOCK_FLASH "\xAA\x13\x00\x00\x00"

#define APDU_UNLOCK "\xAA\x21\x00\x00\x00"

static SCARDCONTEXT pcsc_ctx = 0;
static SCARDHANDLE pcsc_hCard = 0;
static LPSTR pcsc_mszReaders = NULL;

static int pcsc_ctx_open(void)
{
    int ret;
    DWORD dwReaders;

    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;

    ret = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &pcsc_ctx);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardEstablishContext() failed: %08X\n", ret);
        return -1;
    }

    ret = SCardListReaders(pcsc_ctx, NULL, NULL, &dwReaders);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardListReaders() failed: %08X\n", ret);
        return -1;
    }

    pcsc_mszReaders = malloc(sizeof(char) * dwReaders);
    if (pcsc_mszReaders == NULL)
    {
        fprintf(stderr, "malloc: not enough memory\n");
        return -1;
    }

    ret = SCardListReaders(pcsc_ctx, NULL, pcsc_mszReaders, &dwReaders);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardListReaders() failed: %08X\n", ret);
        return -1;
    }

    return 0;
}

static int pcsc_iter_reader(int (*callback)(int index, const char *reader))
{
    int ret;
    LPSTR psReader;

    psReader = pcsc_mszReaders;
    for (int i = 0, n = 0;; i++)
    {
        char *p = pcsc_mszReaders + i;
        if (*p == '\0')
        {
            ret = callback(n, psReader);
            if (ret < 0)
                return -1;
            if (ret > 0)
                return 0;
            if (*(p + 1) == '\0')
            {
                break;
            }
            psReader = p + 1;
            n++;
        }
    }
    return -1;
}

static int pcsc_open_hCard_iter(int index, const char *reader)
{
    int ret;
    int id;
    DWORD dwActiveProtocol;

    id = 0;
    if (getenv("DRIVER_IFID"))
    {
        id = atoi(getenv("DRIVER_IFID"));
    }

    if (id != index)
    {
        return 0;
    }

    ret = SCardConnect(pcsc_ctx, reader, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T0, &pcsc_hCard, &dwActiveProtocol);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardConnect() failed: %08X\n", ret);
        return -1;
    }

    return 1;
}

static int pcsc_open_hCard(void)
{
    return pcsc_iter_reader(pcsc_open_hCard_iter);
}

static void pcsc_close(void)
{
    if (pcsc_mszReaders)
    {
        free(pcsc_mszReaders);
    }

    if (pcsc_hCard)
    {
        SCardDisconnect(pcsc_hCard, SCARD_UNPOWER_CARD);
    }

    if (pcsc_ctx)
    {
        SCardReleaseContext(pcsc_ctx);
    }

    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;
}

static int pcsc_transmit_lowlevel(uint8_t *rx, uint32_t *rx_len, const uint8_t *tx, uint32_t tx_len)
{
    int ret;
    DWORD rx_len_merged;

    rx_len_merged = *rx_len;
    ret = SCardTransmit(pcsc_hCard, SCARD_PCI_T0, tx, tx_len, NULL, rx, &rx_len_merged);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardTransmit() failed: %08X\n", ret);
        return -1;
    }

    *rx_len = rx_len_merged;
    return 0;
}

static int card_poweron(void)
{
    if (pcsc_ctx_open() < 0)
    {
        return -1;
    }
    if (pcsc_open_hCard() < 0)
    {
        return -1;
    }
    return 0;
}

static void card_poweroff(void)
{
    pcsc_close();
}

static void fwupd_enter(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *)APDU_FWUPD_SELECT_AID, sizeof(APDU_FWUPD_SELECT_AID) - 1);

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *)APDU_FWUPD_TRIGGER, sizeof(APDU_FWUPD_TRIGGER) - 1);
}

static int fwupd_unlock_part(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *)APDU_UNLOCK, sizeof(APDU_UNLOCK) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }

    return 0;
}

static int fwupd_block_write(const uint8_t *data)
{
    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    uint32_t block_size = FWUPD_BLOCK_SIZE;

    while (block_size)
    {
        uint8_t tx_size = block_size > FWUPD_TRANS_BLOCK_SIZE ? FWUPD_TRANS_BLOCK_SIZE : block_size;
        uint16_t tx_offset = FWUPD_BLOCK_SIZE - block_size;

        memcpy(pcsc_tx, APDU_BLOCK_WRITE, sizeof(APDU_BLOCK_WRITE) - 1);
        memcpy(pcsc_tx + sizeof(APDU_BLOCK_WRITE) - 1, data + tx_offset, tx_size);

        pcsc_tx[2] = tx_offset >> 8;
        pcsc_tx[3] = tx_offset & 0xFF;
        pcsc_tx[4] = tx_size;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, tx_size + sizeof(APDU_BLOCK_WRITE) - 1);
        if (pcsc_rx[0] != 0x90)
        {
            return -1;
        }

        block_size -= tx_size;
    }

    return 0;
}

static int fwupd_block_verify(const uint8_t *data)
{
    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    uint8_t block_buffer[FWUPD_BLOCK_SIZE];
    uint32_t block_size = FWUPD_BLOCK_SIZE;

    while (block_size)
    {
        uint8_t rx_size = block_size > FWUPD_TRANS_BLOCK_SIZE ? FWUPD_TRANS_BLOCK_SIZE : block_size;
        uint16_t rx_offset = FWUPD_BLOCK_SIZE - block_size;

        memcpy(pcsc_tx, APDU_BLOCK_READ, sizeof(APDU_BLOCK_READ) - 1);
        pcsc_tx[2] = rx_offset >> 8;
        pcsc_tx[3] = rx_offset & 0xFF;
        pcsc_tx[4] = rx_size;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, sizeof(APDU_BLOCK_READ) - 1);
        if (pcsc_rx_len != (rx_size + 2))
        {
            return -1;
        }
        if (pcsc_rx[rx_size] != 0x90)
        {
            return -1;
        }

        memcpy(block_buffer + rx_offset, pcsc_rx, rx_size);

        block_size -= rx_size;
    }

    if (memcmp(block_buffer, data, FWUPD_BLOCK_SIZE) != 0)
    {
        return -1;
    }

    return 0;
}

static int fwupd_block_flash(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *)APDU_BLOCK_FLASH, sizeof(APDU_BLOCK_FLASH) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }

    return 0;
}

static int fwupd_exit(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *)APDU_EXIT, sizeof(APDU_EXIT) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }

    return 0;
}

int fwupd()
{
    FILE *fw;

    fw = fopen("estkme.app.bin", "rb");
    if (fw == NULL)
    {
        printf("estkme.app.bin fopen() failed\r\n");
        return -1;
    }

    printf("init fwupd ");
    fflush(stdout);
    card_poweron();
    card_poweroff();
    if (card_poweron() < 0)
    {
        printf("reset 1 failed\r\n");
        return -1;
    }
    fwupd_enter();
    card_poweroff();
    if (card_poweron() < 0)
    {
        printf("reset 2 failed\r\n");
        return -1;
    }
    printf("done\r\n");

    printf("unlocking ");
    fflush(stdout);
    if (fwupd_unlock_part() < 0)
    {
        printf("failed\r\n");
        return -1;
    }
    printf("done\r\n");

    for (int i = 0; !feof(fw); i++)
    {
        uint8_t block[FWUPD_BLOCK_SIZE];

        switch (fread(block, 1, FWUPD_BLOCK_SIZE, fw))
        {
        case FWUPD_BLOCK_SIZE:
            break;
        case 0:
            continue;
        default:
            return -1;
        }

        printf("program block %d ", i);
        fflush(stdout);

        for (int i = 0;; i++)
        {
            if (i >= 5)
            {
                fprintf(stderr, "send failed\r\n");
                return -1;
            }

            if (fwupd_block_write(block))
            {
                continue;
            }

            if (fwupd_block_verify(block))
            {
                continue;
            }

            break;
        }

        if (fwupd_block_flash())
        {
            fprintf(stderr, "flash failed\r\n");
            return -1;
        }

        printf("done\r\n");
    }

    if (fwupd_exit() < 0)
    {
        fprintf(stderr, "fwupd exit failed\r\n");
        return -1;
    }

    sleep(5);
    card_poweroff();
    printf("fwupd done\r\n");
    return 0;
}

int main(int argc, char *argv[])
{
    int ret;

    ret = fwupd();

    printf("Press any key to continue...\r\n");
    getchar();

    return ret;
}
